"""Ready-to-run commands for common Grafana templating scenarios."""
import datetime
import json
from pathlib import Path
from typing import IO

import click
from grafana_client import GrafanaApi

from grafanarmadillo.alerter import Alerter
from grafanarmadillo.bulk import BulkExporter, BulkImporter
from grafanarmadillo.dashboarder import Dashboarder
from grafanarmadillo.find import Finder
from grafanarmadillo.templator import Templator, make_mapping_templator
from grafanarmadillo.util import load_data


load_file_help = """Should be encoded as json. You can pass this in as a string; or as file using 'file://path/to/file'"""
auto_template_env_help = "The special value '$auto' will automatically provide a value by prepending a '$' to the keys of the grafana mapping"

mapping_help = (
	"Mapping of values to findreplace in the template. "
	+ "The type should be a mapping of environment names (like 'template' or 'production') "
	+ 'to dicts of replacements (for example, `{"region": "SouthWest"}`). '
	+ load_file_help
)
env_grafana_help = "Name of the environment in the mapping file for Grafana, found in the `--mapping` argument"
env_template_help = (
	"Name of the environment in the mapping file for the template, found in the `--mapping` argument. "
	+ auto_template_env_help
)


def make_grafana(config) -> GrafanaApi:
	"""Make a GrafanaApi from the passed config."""
	if isinstance(config.get("auth"), list):
		config["auth"] = tuple(config["auth"])
	return GrafanaApi(**config)


def with_template_options(f):
	"""Add template options to a command."""
	return click.option("--mapping", help=mapping_help)(
		click.option("--env-grafana", help=env_grafana_help)(
			click.option("--env-template", help=env_template_help)(f)
		)
	)


@click.group()
@click.option("--cfg", "-c", help=f"Config for connecting to Grafana. {load_file_help}")
@click.pass_context
def grafanarmadillo(ctx, cfg):
	"""Template Grafana things."""
	ctx.ensure_object(dict)
	if cfg:
		config = load_data(cfg)
		if isinstance(config.get("auth"), list):
			config["auth"] = tuple(config["auth"])
	else:
		config = {}
	ctx.obj["cfg"] = config


@grafanarmadillo.group()
def dashboard():
	"""Manage Grafana dashboards."""


@dashboard.command(name="export")
@click.option("--src", help="Path to the dashboard to capture")
@click.option("--dst", help="Path to write the dashboard to", type=click.File("w"))
@with_template_options
@click.pass_context
def _export_dashboard(ctx, src, dst, mapping, env_grafana, env_template):
	"""Capture a dashboard from Grafana."""
	gfn = make_grafana(ctx.obj["cfg"])

	mapping = load_data(mapping)
	templator = make_mapping_templator(mapping, env_grafana, env_template)
	return export_dashboard(gfn, src, dst, templator)


def export_dashboard(gfn: GrafanaApi, src: str, dst: IO, templator: Templator):
	"""Capture a dashboard from Grafana."""
	finder, dashboarder = Finder(gfn), Dashboarder(gfn)

	dashboard_info = finder.get_from_path(src)
	dashboard_content, _ = dashboarder.export_dashboard(dashboard_info)
	template = templator.make_template_from_dashboard(dashboard_content)

	dst.write(json.dumps(template))


@dashboard.command(name="import")
@click.option("--src", help="Path of the template", type=click.File("r"))
@click.option("--dst", help="Path to write the dashboard to")
@with_template_options
@click.pass_context
def _import_dashboard(ctx, src, dst, mapping, env_grafana, env_template):
	"""Deploy a template to Grafana."""
	gfn = make_grafana(ctx.obj["cfg"])

	mapping = load_data(mapping)
	templator = make_mapping_templator(mapping, env_grafana, env_template)
	return import_dashboard(gfn, src, dst, templator)


def import_dashboard(gfn: GrafanaApi, src: IO, dst: str, templator: Templator):
	"""Deploy a template to Grafana."""
	finder, dashboarder = Finder(gfn), Dashboarder(gfn)

	template = load_data(src.read())

	dashboard_info, folder = finder.create_or_get_dashboard(dst)
	dashboard = templator.make_dashboard_from_template(dashboard_info, template)
	dashboarder.import_dashboard(dashboard, folder)


@grafanarmadillo.group()
def alert():
	"""Manage Grafana alerts."""


@alert.command(name="export")
@click.option("--src", help="Path to the alert to capture")
@click.option("--dst", help="Path to write the alert to", type=click.File("w"))
@with_template_options
@click.pass_context
def _export_alert(ctx, src, dst, mapping, env_grafana, env_template):
	"""Capture an alert from Grafana."""
	gfn = make_grafana(ctx.obj["cfg"])

	mapping = load_data(mapping)
	templator = make_mapping_templator(mapping, env_grafana, env_template)
	return export_alert(gfn, src, dst, templator)


def export_alert(gfn: GrafanaApi, src: str, dst: IO, templator: Templator):
	"""Capture an alert from Grafana."""
	finder, alerter = Finder(gfn), Alerter(gfn)

	alert_info = finder.get_alert_from_path(src)
	alert, _ = alerter.export_alert(alert_info)
	template = templator.make_template_from_dashboard(alert)

	dst.write(json.dumps(template))


@alert.command(name="import")
@click.option("--src", help="Path of the template", type=click.File("r"))
@click.option("--dst", help="Path to write the dashboard to")
@with_template_options
@click.pass_context
def _import_alert(ctx, src, dst, mapping, env_grafana, env_template):
	"""Deploy an alert from a template."""
	gfn = make_grafana(ctx.obj["cfg"])

	mapping = load_data(mapping)
	templator = make_mapping_templator(mapping, env_grafana, env_template)
	return import_alert(gfn, src, dst, templator)


def import_alert(gfn: GrafanaApi, src: IO, dst: str, templator: Templator):
	"""Deploy an alert from a template."""
	finder, alerter = Finder(gfn), Alerter(gfn)

	template = load_data(src.read())

	alert_info, folder_info = finder.create_or_get_alert(dst)
	alert = templator.make_dashboard_from_template(alert_info, template)
	alerter.import_alert(alert, folder_info)


@grafanarmadillo.group()
def migrate():
	"""Migrate between Grafana instances."""


@migrate.command()
@click.option(
	"--grafana-db-path",
	help="Path to the Grafana DB",
	type=click.Path(exists=True, path_type=Path),
)
@click.option(
	"--grafana-container-image",
	help="Grafana image to upgrade to",
	default="grafana/grafana:latest",
)
@click.option(
	"--output-directory",
	"-o",
	help="Path to write output files",
	type=click.Path(exists=True, path_type=Path),
	default=".",
)
@click.option(
	"--grafana-extra-envvars",
	help=f"Environment variables to pass to the Grafana container used to migrate the alerts. {load_file_help}",
	default=None,
)
@click.option(
	"--grafana-migration-timeout",
	help="Timeout in seconds to wait for the Grafana docker container to apply all migrations. A large Grafana instance may take a long time",
	default=300,
	type=int,
)
@click.option(
	"--clone-db",
	help="whether to perform migrations on a clone of the DB or the DB itself",
	type=click.BOOL,
	default=True,
)
@with_template_options
@click.pass_context
def upgrade_alerting(
	ctx,
	grafana_db_path,
	grafana_container_image,
	output_directory,
	grafana_extra_envvars,
	grafana_migration_timeout,
	clone_db,
	mapping,
	env_grafana,
	env_template,
):
	"""
	Migrate from Classic to Unified alerting.

	Migrations from Classic to Unified alerting is done as DB migrations.
	We clone the database and use a docker container to perform the migration.
	Note that because we use a clone of the DB, we need to supply the config with the auth for the live instance.
	"""
	from grafanarmadillo.migrate import migrate

	mapping = load_data(mapping)
	templator = make_mapping_templator(mapping, env_grafana, env_template)

	cfg = ctx.obj["cfg"]
	if grafana_extra_envvars:
		grafana_extra_envvars = load_data(grafana_extra_envvars)
	else:
		grafana_extra_envvars = {}
	migrate(
		cfg,
		grafana_container_image,
		grafana_db_path,
		output_directory,
		templator,
		extra_env_vars=grafana_extra_envvars,
		timeout=datetime.timedelta(seconds=grafana_migration_timeout),
		clone_db=clone_db,
	)


@grafanarmadillo.group()
def resources():
	"""Move many resources to a Grafana."""


@resources.command("import")
@click.option(
	"--root-directory",
	help="Root directory for all resources",
	type=click.Path(exists=True, path_type=Path),
)
@with_template_options
@click.pass_context
def _import_resources(
	ctx,
	root_directory: Path,
	mapping,
	env_grafana,
	env_template,
):
	"""Load exported dashboards and alerts."""
	mapping = load_data(mapping)
	templator = make_mapping_templator(mapping, env_grafana, env_template)

	operator = BulkImporter(ctx.obj["cfg"], root_directory, templator=templator)
	operator.run()


@resources.command("export")
@click.option(
	"--root-directory",
	help="Root directory for all resources",
	type=click.Path(exists=True, path_type=Path),
)
@with_template_options
@click.pass_context
def _export_resources(
	ctx,
	root_directory: Path,
	mapping,
	env_grafana,
	env_template,
):
	"""Export dashboards and alerts from a Grafana instance."""
	mapping = load_data(mapping)
	templator = make_mapping_templator(mapping, env_grafana, env_template)

	operator = BulkExporter(ctx.obj["cfg"], root_directory, templator=templator)
	operator.run()


if __name__ == "__main__":
	grafanarmadillo()



"""Find Grafana dashboards and folders."""

from pathlib import PurePath
from typing import List, Optional, Tuple, Union

from grafana_client import GrafanaApi

from grafanarmadillo.types import (
	AlertSearchResult,
	DashboardSearchResult,
	FolderSearchResult,
)
from grafanarmadillo.util import exactly_one


def _query_message(query_type: str, query: str) -> str:
	"""Format a message detailing the query."""
	return f"type={query_type}, query={query}"


class Finder:
	"""Collection of methods for finding Grafana dashboards and folders."""

	def __init__(self, api: GrafanaApi) -> None:
		super().__init__()
		self.api = api

	def list_dashboards(self) -> List[DashboardSearchResult]:
		"""List all dashboards."""
		return self.api.search.search_dashboards(type_="dash-db")

	def list_alerts(self) -> List[AlertSearchResult]:
		"""List all alerts."""
		return self.api.alertingprovisioning.get_alertrules_all()

	def find_dashboards(self, name: str) -> List[DashboardSearchResult]:
		"""Find all dashboards with a name. Returns exact matches only."""
		return list(
			filter(
				lambda x: x["title"] == name,
				self.api.search.search_dashboards(query=name, type_="dash-db"),
			)
		)

	def _enumerate_dashboards_in_folders(self, folder_ids: List[str]):
		folder_param = ",".join(folder_ids)
		return self.api.search.search_dashboards(
			query=None, type_="dash-db", folder_ids=folder_param
		)

	def get_dashboards_in_folders(self, folder_names: List[str]) -> List[DashboardSearchResult]:
		"""Get all dashboards in folders."""
		folder_objects = list(
			map(lambda folder_name: self.get_folder(name=folder_name), folder_names)
		)
		return self._enumerate_dashboards_in_folders(
			list(map(lambda f: str(f["id"]), folder_objects))
		)

	def get_folder(self, name) -> FolderSearchResult:
		"""Get a folder by name. Folders don't nest, so this will return at most 1 folder."""
		if name == "General":
			return self.api.folder.get_folder_by_id(0)
		else:
			search_result = self.api.search.search_dashboards(query=name, type_="dash-folder")
			return exactly_one(
				list(filter(
					lambda x: x["title"] == name,
					map(lambda sr: self.api.folder.get_folder(sr["uid"]), search_result),
				)),
				_query_message("folder", name),
			)

	def create_or_get_folder(self, name: str) -> FolderSearchResult:
		"""
		Create a new folder if it does not exist.

		Returns the search information if it does.
		"""
		try:
			folder = self.get_folder(name)
		except ValueError:
			folder = self.api.folder.create_folder(name)
		return folder

	def get_dashboard(self, folder_name: str, dashboard_name: str) -> DashboardSearchResult:
		"""
		Get a dashboard by its parent folder and dashboard name.

		Dashboards without a parent are children of the "General" folder.
		"""
		folder_object = self.get_folder(folder_name)
		dashboards = self._enumerate_dashboards_in_folders([str(folder_object["id"])])
		return exactly_one(
			list(filter(lambda d: d["title"] == dashboard_name, dashboards)),
			_query_message("dashboard", f"/{folder_name}/{dashboard_name}"),
		)

	def get_alert(self, folder_name, alert_name) -> AlertSearchResult:
		"""Get an alert by its parent folder and alert name."""
		folder_uid = self.get_folder(folder_name)["uid"]

		return exactly_one(
			list(filter(
				lambda a: a["title"] == alert_name and a["folderUID"] == folder_uid,
				self.api.alertingprovisioning.get_alertrules_all()
			)),
			_query_message("alert", f"/{folder_name}/{alert_name}")
		)

	@staticmethod
	def _resolve_path(path) -> Tuple[str, str]:
		parts = PurePath(path).parts

		# validate path
		if len(parts) > 3:
			# can support maximally ("/", "folder", "dashboard")
			raise ValueError("Dashboard path has too many parts")
		if PurePath(path).is_absolute() or parts[0] == "/" or parts[0] == "\\":
			# removes optional "absolute" signifier
			parts = parts[1:]

		if len(parts) == 2:
			folder, dashboard = parts
		else:
			folder, dashboard = "General", parts[0]

		return folder, dashboard

	def get_from_path(self, path) -> Union[DashboardSearchResult, AlertSearchResult]:
		"""Get a dashboard from a string path like `/folder0/dashboard0`."""
		folder, dashboard = self._resolve_path(path)
		return self.get_dashboard(folder, dashboard)

	def get_alert_from_path(self, path) -> AlertSearchResult:
		"""Get an alert from a string path like `/folder0/alert0`."""
		folder, alert = self._resolve_path(path)
		return self.get_alert(folder, alert)

	def create_or_get_dashboard(self, path: str) -> Tuple[DashboardSearchResult, Optional[FolderSearchResult]]:
		"""
		Create a new empty dashboard if it does not exist.
		
		Returns the search information if it does
		"""
		folder_name, dashboard_name = self._resolve_path(path)

		folder = self.create_or_get_folder(folder_name)

		try:
			dashboard = self.get_dashboard(folder_name, dashboard_name)
		except ValueError:
			self.api.dashboard.update_dashboard(
				{
					"dashboard": {"title": dashboard_name},
					"folderId": folder["id"],
					"folderUid": folder["uid"],
				}
			)
			dashboard = self.get_dashboard(folder_name, dashboard_name)

		return dashboard, folder

	def create_or_get_alert(self, path: str) -> Tuple[AlertSearchResult, FolderSearchResult]:
		"""
		Get the information about an alert or create a new "empty" alert if it does not exist.

		Creating an "empty" alert in Grafana requires filling in a rule.
		We can fake that with a `math` rule that always returns 0.
		"""
		folder_name, alert_name = self._resolve_path(path)

		folder = self.create_or_get_folder(folder_name)

		try:
			alert = self.get_alert(folder_name, alert_name)
		except ValueError:
			self.api.alertingprovisioning.create_alertrule(self._mk_null_alert(folder["uid"], alert_name), disable_provenance=True)
			alert = self.get_alert(folder_name, alert_name)

		return alert, folder

	def _mk_null_alert(self, folder_uid: str, title: str) -> dict:
		"""Fill in the minimum boilerplate for Grafana to let us create an alert."""
		return {
			"title": title,
			"folderUID": folder_uid,
			"condition": "A",
			"ruleGroup": "grafanarmadillo_tmp",
			"data": [
				{
					"refId": "A",
					"relativeTimeRange": {
						"from": 0,
						"to": 0
					},
					"datasourceUid": "__expr__",
					"model": {
						"conditions": [
							{
								"evaluator": {
									"params": [
										0,
										0
									],
									"type": "gt"
								},
								"operator": {
									"type": "and"
								},
								"query": {
									"params": []
								},
								"reducer": {
									"params": [],
									"type": "avg"
								},
								"type": "query"
							}
						],
						"datasource": {
							"name": "Expression",
							"type": "__expr__",
							"uid": "__expr__"
						},
						"expression": "0",
						"intervalMs": 1000000,
						"maxDataPoints": 43200,
						"refId": "A",
						"type": "math"
					}
				}
			],
			"noDataState": "NoData",
			"execErrState": "Error",
			"for": "5m",
			"isPaused": False
		}



"""Make and fill templates for dashboards."""
from typing import Callable, Dict, NewType

from grafanarmadillo.types import (
	DashboardContent,
	DashboardPanel,
	DashboardSearchResult,
)
from grafanarmadillo.util import (
	map_json_strings,
	project_dashboard_identity,
	project_dict,
)


DashboardTransformer = Callable[[DashboardContent], DashboardContent]


def nop(d: DashboardContent) -> DashboardContent:
	"""Pass template through."""
	return d


def findreplace(context: Dict[str, str]) -> DashboardTransformer:
	"""Make DashboardTransformer to make replacements in strings in dashboards."""

	def replace_strings(s: str):
		out = s
		for k, v in context.items():
			out = out.replace(k, v)
		return out

	def _findreplace(d: DashboardContent) -> DashboardContent:
		return map_json_strings(replace_strings, d)

	return _findreplace


def combine_transformers(*transformers: DashboardTransformer) -> DashboardTransformer:
	"""Chain transformers together into one big transformer."""

	def _chained(d: DashboardContent):
		out = d
		for t in transformers:
			out = t(out)
		return out

	return _chained


def panel_transformer(f: Callable[[DashboardPanel], DashboardPanel]) -> DashboardTransformer:
	"""
	Make DashboardTransformer which processes all panels in a dashboard.
	
	Will omit a dashboard if function returns None
	"""

	def _panel_transformer(d: DashboardContent) -> DashboardContent:
		out = d.copy()

		new_panels = list(filter(lambda p: bool(p), (f(p) for p in d["panels"])))

		out["panels"] = new_panels
		return DashboardContent(out)

	return _panel_transformer


class Templator(object):
	"""Collection of methods for filling and making templates."""

	def __init__(
		self,
		make_template: DashboardTransformer = nop,
		fill_template: DashboardTransformer = nop,
	) -> None:
		super().__init__()
		self.make_template = make_template
		self.fill_template = fill_template

	def make_template_from_dashboard(
		self, dashboard: DashboardContent
	) -> DashboardContent:
		"""Convert a dashboard into a one ready for templating."""
		new = dashboard.copy()
		new = project_dict(
			new, {"id", "uid"}, inverse=True
		)  # we don't erase the title so that we can template it later

		return self.make_template(DashboardContent(new))

	def make_dashboard_from_template(
		self, dashboard_info: DashboardSearchResult, template: DashboardContent
	) -> DashboardContent:
		"""Inflate a template."""
		new = template.copy()
		new.update(project_dashboard_identity(dashboard_info))

		return self.fill_template(DashboardContent(new))


EnvMapping = NewType("EnvMapping", Dict[str, Dict[str, str]])
TOK_AUTO_MAPPING = "$auto"


def make_mapping_templator(mapping: EnvMapping, env_grafana: str, env_template: str) -> Templator:
	"""Assemble the templator from the environment mapping."""
	mapping_grafana = mapping[env_grafana]
	if env_template == TOK_AUTO_MAPPING:
		mapping_template = {k: "${%s}" % k for k in mapping_grafana.keys()}
	else:
		mapping_template = mapping[env_template]

	# if some keys in the src mapping are not in the dst mapping
	missing = mapping_grafana.keys() - mapping_template.keys()
	if missing:
		raise ValueError(f"Some keys in the source mapping are not present in the destination mapping. {missing=}")

	grafana_to_template = {v: mapping_template[k] for k, v in mapping_grafana.items()}
	template_to_grafana = {v: k for k, v in grafana_to_template.items()}

	return Templator(make_template=findreplace(grafana_to_template), fill_template=findreplace(template_to_grafana))



"""
Perform bulk operations on Grafana.

BulkOperations defines the basic interface. It is meant to be parametrised in 2 phases:
- information source: Subclasses of BulkOperations implement methods to get orgs and resources
- actions: Subclasses of those classes implement the actions to take

For example:
	BulkGrafanaOperation uses a Grafana instance as its source
	BulkExporter uses BulkGrafanaOperations to list all objects and write them to disk
"""
import logging
from abc import ABC, abstractmethod
from functools import lru_cache
from pathlib import Path
from typing import Generator, List, Tuple

from grafana_client import GrafanaApi

from grafanarmadillo.alerter import Alerter
from grafanarmadillo.dashboarder import Dashboarder
from grafanarmadillo.find import Finder
from grafanarmadillo.templator import Templator
from grafanarmadillo.types import AlertContent, DashboardContent, OrgMeta
from grafanarmadillo.util import exactly_one, read_from_file, write_to_file


l = logging.getLogger(__name__)


@lru_cache
def get_all_orgs(gfn_multiorg) -> List[OrgMeta]:
	"""Get a list of all orgs."""
	return gfn_multiorg.organizations.list_organization()


@lru_cache
def get_org(gfn_multiorg, org_name: str):
	"""Get an org by name, efficiently."""
	return exactly_one(
		list(filter(lambda o: o["name"] == org_name, get_all_orgs(gfn_multiorg))),
		f"org with name {org_name}"
	)


class BulkOperation(ABC):
	"""Run bulk operations on Grafana."""

	def __init__(self, cfg: dict):
		self.cfg = cfg
		self.gfn_multiorg = GrafanaApi(**self.cfg)

	def run(self):
		"""Run this bulk operation."""
		for org, gfn in self.all_orgs():
			for out_path, dashboard_content in self.get_all_dashboards(org, gfn):
				self.each_dashboard(out_path, dashboard_content)
			for out_path, alert_content in self.get_all_alerts(org, gfn):
				self.each_alert(out_path, alert_content)

	@abstractmethod
	def all_orgs(self) -> Generator[Tuple[OrgMeta, GrafanaApi], None, None]:
		"""
		Iterate over all organisations.

		This method should be implemented for each source of information.
		"""

	@abstractmethod
	def get_all_dashboards(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[Path, DashboardContent], None, None]:
		"""
		Iterate over all dashboards.

		This method should be implemented for each source of information.
		"""

	@abstractmethod
	def get_all_alerts(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[Path, AlertContent], None, None]:
		"""
		Iterate over all alerts.

		This method should be implemented for each source of information.
		"""

	@abstractmethod
	def each_dashboard(self, path: Path, dashboard: DashboardContent):
		"""
		Act on each dashboard in Grafana.

		This method should be implemented for each operation.
		"""

	@abstractmethod
	def each_alert(self, path: Path, alert: AlertContent):
		"""
		Act on each alert in Grafana.

		This method should be implemented for each operation.
		"""


class BulkGrafanaOperation(BulkOperation, ABC):
	"""Bulk operations which uses a Grafana instance as its source."""

	def all_orgs(self) -> Generator[Tuple[OrgMeta, GrafanaApi], None, None]:
		"""Iterate over all organisations in Grafana."""
		orgs = self.gfn_multiorg.organizations.list_organization()
		for org in orgs:
			gfn = GrafanaApi(**{**self.cfg, "organization_id": org["id"]})
			yield org, gfn

	def get_all_dashboards(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[Path, DashboardContent], None, None]:
		"""Get all dashboards."""
		finder, dashboarder = Finder(gfn), Dashboarder(gfn)
		dashboards = finder.list_dashboards()
		for dashboard in dashboards:
			dashboard_content, folder = dashboarder.export_dashboard(dashboard)

			if folder is None:
				folder_name = "General"
			else:
				folder_name = folder["name"]

			out_path = Path(org["name"], folder_name, dashboard_content["title"])

			yield out_path, dashboard_content

	def get_all_alerts(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[Path, AlertContent], None, None]:
		"""Get all alerts."""
		finder, alerter = Finder(gfn), Alerter(gfn)
		alerts = finder.list_alerts()
		for alert in alerts:
			alert_content, folder = alerter.export_alert(alert)

			folder_name = folder["title"]
			out_path = Path(org["name"], folder_name, alert_content["title"])

			yield out_path, alert_content


class BulkFileOperation(BulkOperation, ABC):
	"""Bulk operation which uses a filetree as its source."""

	def __init__(self, cfg: dict, root_directory: Path):
		self.root_directory = root_directory
		super().__init__(cfg)

	def all_orgs(self) -> Generator[Tuple[OrgMeta, GrafanaApi], None, None]:
		"""Iterate over all organisations in Grafana."""
		orgs = {o.name for o in self.root_directory.glob("*/*")}
		for org_name in orgs:
			org = self.gfn_multiorg.organization.find_organization(org_name)
			gfn = GrafanaApi(**{**self.cfg, "organization_id": org["id"]})

			yield org, gfn

	def get_all_dashboards(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[Path, DashboardContent], None, None]:
		"""Get all dashboards."""
		folders = (self.root_directory / "dashboards" / org["name"]).glob("*")
		for folder_path in folders:
			for dashboard_path in folder_path.glob("*.json"):
				content = read_from_file(dashboard_path)
				yield Path(org["name"], folder_path.name, dashboard_path.name), content

	def get_all_alerts(self, org: OrgMeta, gfn: GrafanaApi) -> Generator[Tuple[Path, AlertContent], None, None]:
		"""Get all alerts."""
		folders = (self.root_directory / "alerts" / org["name"]).glob("*")
		for folder_path in folders:
			for alert_path in folder_path.glob("*.json"):
				content = read_from_file(alert_path)
				yield Path(org["name"], folder_path.name, alert_path.name), content


class BulkExporter(BulkGrafanaOperation):
	"""Export all resources from Grafana to files."""

	def __init__(self, cfg: dict, root_directory: Path, templator: Templator):
		self.root_directory = root_directory
		self.templator = templator
		super().__init__(cfg)

	def each_dashboard(self, path: Path, dashboard: DashboardContent):
		"""Write each dashboard to files."""
		dashboard_templated = self.templator.make_template_from_dashboard(dashboard)
		l.info(f"export dashboard path={path}")
		write_to_file((self.root_directory / "dashboards" / path).with_suffix(".json"), dashboard_templated)

	def each_alert(self, path: Path, alert: AlertContent):
		"""Write each alert to files."""
		alert_templated = self.templator.make_template_from_dashboard(alert)
		l.info(f"export alert path={path}")
		write_to_file((self.root_directory / "alerts" / path).with_suffix(".json"), alert_templated)


class BulkImporter(BulkFileOperation):
	"""Import all resources from files into Grafana."""

	def __init__(self, cfg: dict, root_directory: Path, templator: Templator):
		self.templator = templator
		super().__init__(cfg, root_directory)

	def each_dashboard(self, path: Path, dashboard: DashboardContent):
		"""Import each dashboard into Grafana."""
		org_name, folder_name, dashboard_name = path.parts

		org = get_org(self.gfn_multiorg, org_name)
		gfn = GrafanaApi(**{**self.cfg, "organization_id": org["id"]})

		finder, dashboarder = Finder(gfn), Dashboarder(gfn)
		folder = finder.create_or_get_folder(folder_name)
		dashboard_templated = self.templator.make_dashboard_from_template(
			dashboard, dashboard
		)
		l.info(f"import dashboard path={path}")
		dashboarder.import_dashboard(dashboard_templated, folder)

	def each_alert(self, path: Path, alert: AlertContent):
		"""Import each alert into Grafana."""
		org_name, folder_name, dashboard_name = path.parts

		org = get_org(self.gfn_multiorg, org_name)
		gfn = GrafanaApi(**{**self.cfg, "organization_id": org["id"]})

		finder, alerter = Finder(gfn), Alerter(gfn)
		folder = finder.create_or_get_folder(folder_name)
		alert_templated = self.templator.make_dashboard_from_template(alert, alert)
		l.info(f"import alert path={path}")
		alerter.import_alert(alert_templated, folder)



# -*- coding: utf-8 -*-
"""Config file for Sphinx-docs."""
from __future__ import unicode_literals

import os

import sphinx_py3doc_enhanced_theme


extensions = [
	"sphinx.ext.autodoc",
	"sphinx.ext.autosummary",
	"sphinx.ext.coverage",
	"sphinx.ext.doctest",
	"sphinx.ext.extlinks",
	"sphinx.ext.ifconfig",
	"sphinx.ext.napoleon",
	"sphinx.ext.todo",
	"sphinx.ext.viewcode",
	"sphinx.ext.autosectionlabel",
]

todo_include_todos = True

exclude_patterns = [
	"i_*",
]

if os.getenv("SPELLCHECK"):
	extensions += ("sphinxcontrib.spelling",)
	spelling_show_suggestions = True
	spelling_lang = "en_US"
	# https://sphinxcontrib-spelling.readthedocs.io/en/latest/customize.html
	spelling_word_list_filename = ["../spelling_wordlist.txt"]

source_suffix = ".rst"
master_doc = "index"
project = "grafanarmadillo"
year = "2021"
author = "lilatomic"
copyright = "{0}, {1}".format(year, author)
version = release = "0.5.0"

pygments_style = "trac"
templates_path = ["."]
extlinks = {
	"issue": ("https://github.com/lilatomic/grafanarmadillo/issues/%s", "#"),  # noqa: E501
	"pr": ("https://github.com/lilatomic/grafanarmadillo/pull/%s", "PR #"),  # noqa: E501
}

# codecov io closes connection if host is accessed too repetitively.
# codecov links are ignored here for the same reason there's a sleep
# in the .travis.yml file
# see https://github.com/codecov/codecov-python/issues/158
linkcheck_ignore = [
	r"https://codecov.io/gh/lilatomic/grafanarmadillo/*",
]

html_theme = "sphinx_py3doc_enhanced_theme"
html_theme_path = [sphinx_py3doc_enhanced_theme.get_html_theme_path()]
html_theme_options = {
	"githuburl": "https://github.com/lilatomic/grafanarmadillo",
}

html_use_smartypants = True
html_last_updated_fmt = "%b %d, %Y"
html_split_index = False
html_sidebars = {
	"**": ["searchbox.html", "globaltoc.html", "sourcelink.html"],
}
html_short_title = "%s-%s" % (project, version)

napoleon_use_ivar = True
napoleon_use_rtype = False
napoleon_use_param = False
